/*
Copyright (C) 2006 - 2011 Evan Teran
                          eteran@alum.rit.edu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "DebuggerCore.h"
#include "DebugEvent.h"
#include "State.h"
#include "PlatformState.h"
#include <QDebug>
#include <QStringList>
#include <windows.h>


class Win32Handle {
public:
	Win32Handle() : handle(NULL) {}
	Win32Handle(HANDLE handle) : handle(handle) {}
	~Win32Handle() { if(handle) { CloseHandle(handle); } }
private:
	Win32Handle(const Win32Handle&);
	Win32Handle &operator=(const Win32Handle&);

public:
	HANDLE get() const { return handle; }

public:
	operator void*() const { return reinterpret_cast<void*>(handle != 0); }

private:
	HANDLE handle;
};

//------------------------------------------------------------------------------
// Name: DebuggerCore()
// Desc: constructor
//------------------------------------------------------------------------------
DebuggerCore::DebuggerCore() : page_size_(0), process_handle_(0), start_address(0), image_base(0) {
	DebugSetProcessKillOnExit(false);

	SYSTEM_INFO sys_info;
	GetSystemInfo(&sys_info);
	page_size_ = sys_info.dwPageSize;
}

//------------------------------------------------------------------------------
// Name: ~DebuggerCore()
// Desc:
//------------------------------------------------------------------------------
DebuggerCore::~DebuggerCore() {
	detach();
}

//------------------------------------------------------------------------------
// Name: page_size() const
// Desc: returns the size of a page on this system
//------------------------------------------------------------------------------
edb::address_t DebuggerCore::page_size() const {
	return page_size_;
}

//------------------------------------------------------------------------------
// Name: wait_debug_event(DebugEvent &event, bool &ok, int secs)
// Desc: waits for a debug event, secs is a timeout (but is not yet respected)
//       ok will be set to false if the timeout expires
//------------------------------------------------------------------------------
bool DebuggerCore::wait_debug_event(DebugEvent &event, int msecs) {
	if(attached()) {
		DEBUG_EVENT de;
		while(WaitForDebugEvent(&de, msecs == 0 ? INFINITE : msecs) != 0) {
			switch(de.dwDebugEventCode) {
			case CREATE_THREAD_DEBUG_EVENT:
				threads_.insert(de.dwThreadId);
				ContinueDebugEvent(de.dwProcessId, de.dwThreadId, DBG_CONTINUE);
				break;
			case EXIT_THREAD_DEBUG_EVENT:
				threads_.remove(de.dwThreadId);
				ContinueDebugEvent(de.dwProcessId, de.dwThreadId, DBG_CONTINUE);
				break;
			case CREATE_PROCESS_DEBUG_EVENT:
				start_address	= reinterpret_cast<edb::address_t>(de.u.CreateProcessInfo.lpStartAddress);
				image_base		= reinterpret_cast<edb::address_t>(de.u.CreateProcessInfo.lpBaseOfImage);
				ContinueDebugEvent(de.dwProcessId, de.dwThreadId, DBG_CONTINUE);
				break;
			case EXIT_PROCESS_DEBUG_EVENT:
				start_address	= 0;
				image_base		= 0;
				ContinueDebugEvent(de.dwProcessId, de.dwThreadId, DBG_CONTINUE);
				break;
			case LOAD_DLL_DEBUG_EVENT:
			case UNLOAD_DLL_DEBUG_EVENT:
			case OUTPUT_DEBUG_STRING_EVENT:
			case RIP_EVENT:
				ContinueDebugEvent(de.dwProcessId, de.dwThreadId, DBG_CONTINUE);
				break;
			case EXCEPTION_DEBUG_EVENT:
				event = DebugEvent(de);
				active_thread_ = event.thread();
				return true;
			}
		}
	}
	return false;
}

//------------------------------------------------------------------------------
// Name: read_pages(edb::address_t address, void *buf, std::size_t count)
// Desc: reads <count> pages from the process starting at <address>
// Note: buf's size must be >= count * page_size()
// Note: address should be page aligned.
//------------------------------------------------------------------------------
bool DebuggerCore::read_pages(edb::address_t address, void *buf, std::size_t count) {

	Q_ASSERT(buf != 0);

	bool ok = false;

	if(attached()) {
		const edb::address_t orig_address	= address;
		const edb::address_t end_address	= orig_address + page_size() * count;

		if(process_handle_ != 0) {
			SIZE_T bytes_read;
			ok = ReadProcessMemory(
				process_handle_,
				reinterpret_cast<LPVOID>(address),
				buf,
				count * page_size(),
				&bytes_read);
		}

		if(ok) {
			quint8 *const orig_ptr = reinterpret_cast<quint8 *>(buf);
			// TODO: handle if breakponts have a size more than 1!
			Q_FOREACH(const QSharedPointer<Breakpoint> &bp, breakpoints_) {
				if(bp->address() >= orig_address && bp->address() < end_address) {
					// show the original bytes in the buffer..
					orig_ptr[bp->address() - orig_address] = bp->original_bytes()[0];
				}
			}
		}
	}
	return ok;
}

//------------------------------------------------------------------------------
// Name: read_bytes(edb::address_t address, void *buf, std::size_t len)
// Desc: reads <len> bytes into <buf> starting at <address>
// Note: if the read failed, the part of the buffer that could not be read will
//       be filled with 0xff bytes
//------------------------------------------------------------------------------
bool DebuggerCore::read_bytes(edb::address_t address, void *buf, std::size_t len) {

	Q_ASSERT(buf != 0);

	bool ok = false;

	if(attached()) {
		const edb::address_t orig_address	= address;
		const edb::address_t end_address	= orig_address + len;

		if(process_handle_ != 0) {
			SIZE_T bytes_read;
			ok = ReadProcessMemory(
				process_handle_,
				reinterpret_cast<LPVOID>(address),
				buf,
				len,
				&bytes_read);
		}
		// TODO: do the documented 0xff fill

		if(ok) {
			quint8 *const orig_ptr = reinterpret_cast<quint8 *>(buf);
			// TODO: handle if breakponts have a size more than 1!
			Q_FOREACH(const QSharedPointer<Breakpoint> &bp, breakpoints_) {
				if(bp->address() >= orig_address && bp->address() < end_address) {
					// show the original bytes in the buffer..
					orig_ptr[bp->address() - orig_address] = bp->original_bytes()[0];
				}
			}
		}
	}
	return ok;
}

//------------------------------------------------------------------------------
// Name: write_bytes(edb::address_t address, const void *buf, std::size_t len)
// Desc: writes <len> bytes from <buf> starting at <address>
//------------------------------------------------------------------------------
bool DebuggerCore::write_bytes(edb::address_t address, const void *buf, std::size_t len) {

	Q_ASSERT(buf != 0);

	bool ok = false;
	if(attached()) {
		if(process_handle_ != 0) {
			SIZE_T bytes_written;
			ok = WriteProcessMemory(
				process_handle_,
				reinterpret_cast<LPVOID>(address),
				buf,
				len,
				&bytes_written);

			if(ok) {
				FlushInstructionCache(process_handle_, reinterpret_cast<LPVOID>(address), bytes_written);
			}
		}
	}
	return ok;
}

//------------------------------------------------------------------------------
// Name: attach(edb::pid_t pid)
// Desc:
//------------------------------------------------------------------------------
bool DebuggerCore::attach(edb::pid_t pid) {
	detach();

	if(DebugActiveProcess(pid)) {
		DEBUG_EVENT ev;
		if(WaitForDebugEvent(&ev, INFINITE)) {

			Q_ASSERT(ev.dwDebugEventCode == CREATE_PROCESS_DEBUG_EVENT);

			// close the image file handle
			CloseHandle(ev.u.CreateProcessInfo.hFile);

			// remember the process handle
			process_handle_ = ev.u.CreateProcessInfo.hProcess;

			// for now we'l close the thread's handle
			CloseHandle(ev.u.CreateProcessInfo.hThread);

			pid_ = pid;
			active_thread_ = ev.dwThreadId;

			// tell it to continue to we can catch the rest of the events
			ContinueDebugEvent(ev.dwProcessId, ev.dwThreadId, DBG_CONTINUE);

			return true;
		}
	}

	return false;
}

//------------------------------------------------------------------------------
// Name: detach()
// Desc:
//------------------------------------------------------------------------------
void DebuggerCore::detach() {
	if(attached()) {
		clear_breakpoints();
		threads_.clear();

		// why is this illegal?
		//CloseHandle(process_handle_);
		process_handle_ = 0;

		DebugActiveProcessStop(pid());
		pid_ = 0;
	}
}

//------------------------------------------------------------------------------
// Name: kill()
// Desc:
//------------------------------------------------------------------------------
void DebuggerCore::kill() {
	if(attached()) {
		clear_breakpoints();
		if(process_handle_ != 0) {
			TerminateProcess(process_handle_, -1);
			CloseHandle(process_handle_);
			process_handle_ = 0;
		}
		pid_ = 0;
	}
}

//------------------------------------------------------------------------------
// Name: pause()
// Desc: stops *all* threads of a process
//------------------------------------------------------------------------------
void DebuggerCore::pause() {

	if(attached()) {
		if(process_handle_ != 0) {
			DebugBreakProcess(process_handle_);
		}
	}
}

//------------------------------------------------------------------------------
// Name: resume(edb::EVENT_STATUS status)
// Desc:
//------------------------------------------------------------------------------
void DebuggerCore::resume(edb::EVENT_STATUS status) {
	// TOOD: assert that we are paused

	if(attached()) {
		if(status != edb::DEBUG_STOP) {
			// TODO: does this resume *all* threads?
			ContinueDebugEvent(
				pid(),
				active_thread(),
				(status == edb::DEBUG_CONTINUE) ? (DBG_CONTINUE) : (DBG_EXCEPTION_NOT_HANDLED));
		}
	}
}

//------------------------------------------------------------------------------
// Name: step(edb::EVENT_STATUS status)
// Desc:
//------------------------------------------------------------------------------
void DebuggerCore::step(edb::EVENT_STATUS status) {
	// TOOD: assert that we are paused

	if(attached()) {
		if(status != edb::DEBUG_STOP) {
			if(const Win32Handle th = OpenThread(THREAD_ALL_ACCESS, FALSE, active_thread())) {
				CONTEXT	context;
				context.ContextFlags = CONTEXT_CONTROL;
				GetThreadContext(th.get(), &context);

				context.EFlags |= (1 << 8); // set the TF bit
				SetThreadContext(th.get(), &context);

				ContinueDebugEvent(
					pid(),
					active_thread(),
					(status == edb::DEBUG_CONTINUE) ? (DBG_CONTINUE) : (DBG_EXCEPTION_NOT_HANDLED));
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: get_state(State &state)
// Desc:
//------------------------------------------------------------------------------
void DebuggerCore::get_state(State &state) {
	// TOOD: assert that we are paused

	PlatformState *state_impl = static_cast<PlatformState *>(state.impl_);

	if(attached() && state_impl) {
		if(const Win32Handle th = OpenThread(THREAD_ALL_ACCESS, FALSE, active_thread())) {
			state_impl->context_.ContextFlags = CONTEXT_FULL | CONTEXT_DEBUG_REGISTERS | CONTEXT_FLOATING_POINT;
			GetThreadContext(th.get(), &state_impl->context_);

			LDT_ENTRY ldt_entry;
			if(GetThreadSelectorEntry(th.get(), state_impl->context_.SegGs, &ldt_entry)) {
				state_impl->gs_base_ = ldt_entry.BaseLow | (ldt_entry.HighWord.Bits.BaseMid << 16) | (ldt_entry.HighWord.Bits.BaseHi << 24);;
			}

			if(GetThreadSelectorEntry(th.get(), state_impl->context_.SegFs, &ldt_entry)) {
				state_impl->fs_base_ = ldt_entry.BaseLow | (ldt_entry.HighWord.Bits.BaseMid << 16) | (ldt_entry.HighWord.Bits.BaseHi << 24);;
			}
		}
	} else {
		state.clear();
	}
}

//------------------------------------------------------------------------------
// Name: set_state(const State &state)
// Desc:
//------------------------------------------------------------------------------
void DebuggerCore::set_state(const State &state) {

	// TOOD: assert that we are paused

	PlatformState *state_impl = static_cast<PlatformState *>(state.impl_);

	if(attached()) {
		state_impl->context_.ContextFlags = CONTEXT_FULL | CONTEXT_DEBUG_REGISTERS | CONTEXT_FLOATING_POINT;

		if(const Win32Handle th = OpenThread(THREAD_ALL_ACCESS, FALSE, active_thread())) {
			SetThreadContext(th.get(), &state_impl->context_);
		}
	}
}

//------------------------------------------------------------------------------
// Name: open(const QString &path, const QString &cwd, const QStringList &args, const QString &tty)
// Desc:
//------------------------------------------------------------------------------
bool DebuggerCore::open(const QString &path, const QString &cwd, const QStringList &args, const QString &tty) {

	Q_UNUSED(path);
	Q_UNUSED(cwd);
	Q_UNUSED(args);
	Q_UNUSED(tty);

	detach();

	STARTUPINFO			startup_info = {};
	PROCESS_INFORMATION	process_info = {};

	LPVOID env = (LPVOID)GetEnvironmentStringsW();
	WCHAR command_path[MAX_PATH];

	wcscpy_s(command_path, path.utf16());

	if(CreateProcess(
			0,		// exe
			command_path,		// commandline
			NULL,	// default security attributes
			NULL,	// default thread security too
			FALSE,	// inherit handles
			CREATE_SUSPENDED | DEBUG_PROCESS | DEBUG_ONLY_THIS_PROCESS | CREATE_UNICODE_ENVIRONMENT | CREATE_NEW_CONSOLE,
			env,	// environment
			cwd.utf16(),		// cwd
			&startup_info,
			&process_info)) {

		pid_			= process_info.dwProcessId;
		active_thread_	= process_info.dwThreadId;

		ResumeThread(process_info.hThread);

		process_handle_ = process_info.hProcess;
		CloseHandle(process_info.hThread);
		return true;
	}

	return false;
}

//------------------------------------------------------------------------------
// Name: create_state()
// Desc:
//------------------------------------------------------------------------------
StateInterface *DebuggerCore::create_state() const {
	return new PlatformState;
}

Q_EXPORT_PLUGIN2(DebuggerCore, DebuggerCore)
