/*
Copyright (C) 2006 - 2011 Evan Teran
                          eteran@alum.rit.edu

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "PlatformState.h"

//------------------------------------------------------------------------------
// Name: PlatformState()
// Desc:
//------------------------------------------------------------------------------
PlatformState::PlatformState() {
	memset(&context_, 0, sizeof(context_));
	fs_base_ = 0;
	gs_base_ = 0;
}

//------------------------------------------------------------------------------
// Name: PlatformState::copy() const
// Desc: makes a copy of the state object
//------------------------------------------------------------------------------
StateInterface *PlatformState::copy() const {
	return new PlatformState(*this);
}

//------------------------------------------------------------------------------
// Name: flags_to_string(edb::reg_t flags) const
// Desc: returns the flags in a string form appropriate for this platform
//------------------------------------------------------------------------------
QString PlatformState::flags_to_string(edb::reg_t flags) const {
	char buf[14];
	qsnprintf(
		buf,
		sizeof(buf),
		"%c %c %c %c %c %c %c",
		((flags & 0x001) ? 'C' : 'c'),
		((flags & 0x004) ? 'P' : 'p'),
		((flags & 0x010) ? 'A' : 'a'),
		((flags & 0x040) ? 'Z' : 'z'),
		((flags & 0x080) ? 'S' : 's'),
		((flags & 0x400) ? 'D' : 'd'),
		((flags & 0x800) ? 'O' : 'o'));

	return buf;
}

//------------------------------------------------------------------------------
// Name: flags_to_string() const
// Desc: returns the flags in a string form appropriate for this platform
//------------------------------------------------------------------------------
QString PlatformState::flags_to_string() const {
	return flags_to_string(flags());
}

//------------------------------------------------------------------------------
// Name: value(const QString &reg) const
// Desc: returns a Register object which represents the register with the name
//       supplied
//------------------------------------------------------------------------------
Register PlatformState::value(const QString &reg) const {
	const QString lreg = reg.toLower();

#if defined(EDB_X86)
	if(lreg == "eax")			return Register("eax", context_.Eax, Register::TYPE_GPR);
	else if(lreg == "ebx")		return Register("ebx", context_.Ebx, Register::TYPE_GPR);
	else if(lreg == "ecx")		return Register("ecx", context_.Ecx, Register::TYPE_GPR);
	else if(lreg == "edx")		return Register("edx", context_.Edx, Register::TYPE_GPR);
	else if(lreg == "ebp")		return Register("ebp", context_.Ebp, Register::TYPE_GPR);
	else if(lreg == "esp")		return Register("esp", context_.Esp, Register::TYPE_GPR);
	else if(lreg == "esi")		return Register("esi", context_.Esi, Register::TYPE_GPR);
	else if(lreg == "edi")		return Register("edi", context_.Edi, Register::TYPE_GPR);
	else if(lreg == "eip")		return Register("eip", context_.Eip, Register::TYPE_IP);
	else if(lreg == "ax")		return Register("ax", context_.Eax & 0xffff, Register::TYPE_GPR);
	else if(lreg == "bx")		return Register("bx", context_.Ebx & 0xffff, Register::TYPE_GPR);
	else if(lreg == "cx")		return Register("cx", context_.Ecx & 0xffff, Register::TYPE_GPR);
	else if(lreg == "dx")		return Register("dx", context_.Edx & 0xffff, Register::TYPE_GPR);
	else if(lreg == "bp")		return Register("bp", context_.Ebp & 0xffff, Register::TYPE_GPR);
	else if(lreg == "sp")		return Register("sp", context_.Esp & 0xffff, Register::TYPE_GPR);
	else if(lreg == "si")		return Register("si", context_.Esi & 0xffff, Register::TYPE_GPR);
	else if(lreg == "di")		return Register("di", context_.Edi & 0xffff, Register::TYPE_GPR);
	else if(lreg == "al")		return Register("al", context_.Eax & 0xff, Register::TYPE_GPR);
	else if(lreg == "bl")		return Register("bl", context_.Ebx & 0xff, Register::TYPE_GPR);
	else if(lreg == "cl")		return Register("cl", context_.Ecx & 0xff, Register::TYPE_GPR);
	else if(lreg == "dl")		return Register("dl", context_.Edx & 0xff, Register::TYPE_GPR);
	else if(lreg == "ah")		return Register("ah", (context_.Eax >> 8) & 0xff, Register::TYPE_GPR);
	else if(lreg == "bh")		return Register("bh", (context_.Ebx >> 8) & 0xff, Register::TYPE_GPR);
	else if(lreg == "ch")		return Register("ch", (context_.Ecx >> 8) & 0xff, Register::TYPE_GPR);
	else if(lreg == "dh")		return Register("dh", (context_.Edx >> 8) & 0xff, Register::TYPE_GPR);
	else if(lreg == "cs")		return Register("cs", context_.SegCs, Register::TYPE_SEG, 0);
	else if(lreg == "ds")		return Register("ds", context_.SegDs, Register::TYPE_SEG, 0);
	else if(lreg == "es")		return Register("es", context_.SegEs, Register::TYPE_SEG, 0);
	else if(lreg == "fs")		return Register("fs", context_.SegFs, Register::TYPE_SEG, fs_base_);
	else if(lreg == "gs")		return Register("gs", context_.SegGs, Register::TYPE_SEG, gs_base_);
	else if(lreg == "ss") 		return Register("ss", context_.SegSs, Register::TYPE_SEG, 0);
	else if(lreg == "eflags") 	return Register("eflags", context_.EFlags, Register::TYPE_COND);
#elif defined(EDB_X86_64)
#endif

	return Register();
}

//------------------------------------------------------------------------------
// Name: frame_pointer() const
// Desc: returns what is conceptually the frame pointer for this platform
//------------------------------------------------------------------------------
edb::address_t PlatformState::frame_pointer() const {
#if defined(EDB_X86)
	return context_.Ebp;
#elif defined(EDB_X86_64)
#endif
}

//------------------------------------------------------------------------------
// Name: instruction_pointer() const
// Desc: returns the instruction pointer for this platform
//------------------------------------------------------------------------------
edb::address_t PlatformState::instruction_pointer() const {
#if defined(EDB_X86)
	return context_.Eip;
#elif defined(EDB_X86_64)
#endif
}

//------------------------------------------------------------------------------
// Name: stack_pointer() const
// Desc: returns the stack pointer for this platform
//------------------------------------------------------------------------------
edb::address_t PlatformState::stack_pointer() const {
#if defined(EDB_X86)
	return context_.Esp;
#elif defined(EDB_X86_64)
#endif
}

//------------------------------------------------------------------------------
// Name: debug_register(int n) const
// Desc:
//------------------------------------------------------------------------------
edb::reg_t PlatformState::debug_register(int n) const {

	switch(n) {
	case 0: return context_.Dr0;
	case 1: return context_.Dr1;
	case 2: return context_.Dr2;
	case 3: return context_.Dr3;
	case 6: return context_.Dr6;
	case 7: return context_.Dr7;
	}
	return 0;
}

//------------------------------------------------------------------------------
// Name: flags() const
// Desc:
//------------------------------------------------------------------------------
edb::reg_t PlatformState::flags() const {
#if defined(EDB_X86)
	return context_.EFlags;
#elif defined(EDB_X86_64)
#endif
}

//------------------------------------------------------------------------------
// Name: fpu_register(int n) const
// Desc:
//------------------------------------------------------------------------------
long double PlatformState::fpu_register(int n) const {
	Q_UNUSED(n);
	return 0.0;
}

//------------------------------------------------------------------------------
// Name: adjust_stack(int bytes)
// Desc:
//------------------------------------------------------------------------------
void PlatformState::adjust_stack(int bytes) {
#if defined(EDB_X86)
	context_.Esp += bytes;
#elif defined(EDB_X86_64)
#endif
}

//------------------------------------------------------------------------------
// Name: clear()
// Desc:
//------------------------------------------------------------------------------
void PlatformState::clear() {
	memset(&context_, 0, sizeof(context_));
	fs_base_ = 0;
	gs_base_ = 0;
}

//------------------------------------------------------------------------------
// Name: set_debug_register(int n, edb::reg_t value)
// Desc:
//------------------------------------------------------------------------------
void PlatformState::set_debug_register(int n, edb::reg_t value) {
	switch(n) {
	case 0: context_.Dr0 = value; break;
	case 1: context_.Dr1 = value; break;
	case 2: context_.Dr2 = value; break;
	case 3: context_.Dr3 = value; break;
	case 6: context_.Dr6 = value; break;
	case 7: context_.Dr7 = value; break;
	default:
		break;
	}
}

//------------------------------------------------------------------------------
// Name: set_flags(edb::reg_t flags)
// Desc:
//------------------------------------------------------------------------------
void PlatformState::set_flags(edb::reg_t flags) {
#if defined(EDB_X86)
	context_.EFlags = flags;
#elif defined(EDB_X86_64)
#endif
}

//------------------------------------------------------------------------------
// Name: set_instruction_pointer(edb::address_t value)
// Desc:
//------------------------------------------------------------------------------
void PlatformState::set_instruction_pointer(edb::address_t value) {
#if defined(EDB_X86)
	context_.Eip = value;
#elif defined(EDB_X86_64)
#endif
}

//------------------------------------------------------------------------------
// Name: set_register(const QString &name, edb::reg_t value)
// Desc:
//------------------------------------------------------------------------------
void PlatformState::set_register(const QString &name, edb::reg_t value) {

	const QString lreg = name.toLower();
#if defined(EDB_X86)
	if(lreg == "eax") { context_.Eax = value; }
	else if(lreg == "ebx") { context_.Ebx = value; }
	else if(lreg == "ecx") { context_.Ecx = value; }
	else if(lreg == "edx") { context_.Edx = value; }
	else if(lreg == "ebp") { context_.Ebp = value; }
	else if(lreg == "esp") { context_.Esp = value; }
	else if(lreg == "esi") { context_.Esi = value; }
	else if(lreg == "edi") { context_.Edi = value; }
	else if(lreg == "eip") { context_.Eip = value; }
	else if(lreg == "cs") { context_.SegCs = value; }
	else if(lreg == "ds") { context_.SegDs = value; }
	else if(lreg == "es") { context_.SegEs = value; }
	else if(lreg == "fs") { context_.SegFs = value; }
	else if(lreg == "gs") { context_.SegGs = value; }
	else if(lreg == "ss") { context_.SegSs = value; }
	else if(lreg == "eflags") { context_.EFlags = value; }
#elif defined(EDB_X86_64)
#endif
}


